import os
import shutil
from logging import getLogger
from pathlib import Path
from typing import Tuple

from PyQt5.QtCore import pyqtSignal, QRunnable, QObject, Qt
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QLabel, QFileDialog
from legendary.models.game import VerifyResult, InstalledGame
from legendary.lfs.utils import validate_files

from rare.shared import LegendaryCoreSingleton
from rare.utils.extra_widgets import PathEdit

logger = getLogger("MoveGame")


class MoveGamePopUp(QWidget):
    move_clicked = pyqtSignal(str)
    browse_done = pyqtSignal()

    def __init__(self):
        super(MoveGamePopUp, self).__init__()
        layout: QVBoxLayout = QVBoxLayout()
        self.install_path = str()
        self.core = LegendaryCoreSingleton()
        self.move_path_edit = PathEdit(str(), QFileDialog.Directory, edit_func=self.edit_func_move_game)
        self.move_path_edit.path_select.clicked.connect(self.emit_browse_done_signal)

        self.move_game = QPushButton(self.tr("Move"))
        self.move_game.setMaximumWidth(50)
        self.move_game.clicked.connect(self.emit_move_game_signal)

        self.warn_overwriting = QLabel()

        middle_layout = QHBoxLayout()
        middle_layout.setAlignment(Qt.AlignRight)
        middle_layout.addWidget(self.warn_overwriting, stretch=1)
        middle_layout.addWidget(self.move_game)

        bottom_layout = QVBoxLayout()
        self.aval_space_label = QLabel()
        self.req_space_label = QLabel()
        bottom_layout.addWidget(self.aval_space_label)
        bottom_layout.addWidget(self.req_space_label)

        layout.addWidget(self.move_path_edit)
        layout.addLayout(middle_layout)
        layout.addLayout(bottom_layout)

        self.setLayout(layout)

    def emit_move_game_signal(self):
        self.move_clicked.emit(self.move_path_edit.text())

    def emit_browse_done_signal(self):
        self.browse_done.emit()

    def refresh_indicator(self):
        # needed so the edit_func gets run again
        text = self.move_path_edit.text()
        self.move_path_edit.setText(str())
        self.move_path_edit.setText(text)

    @staticmethod
    def is_different_drive(dir1: str, dir2: str):
        return os.stat(dir1).st_dev != os.stat(dir2).st_dev

    def edit_func_move_game(self, dir_selected):
        self.move_game.setEnabled(True)
        self.warn_overwriting.setHidden(True)

        def helper_func(reason: str) -> Tuple[bool, str, str]:
            self.move_game.setEnabled(False)
            return False, dir_selected, self.tr(reason)

        if not self.install_path or not dir_selected:
            return helper_func("You need to provide a directory.")

        install_path = Path(self.install_path).resolve()
        dest_path = Path(dir_selected).resolve()
        dest_path_with_suffix = dest_path.joinpath(install_path.stem).resolve()

        if not dest_path.is_dir():
            return helper_func("Directory doesn't exist or file selected.")

        # Get free space on drive and size of game folder
        _, _, free_space = shutil.disk_usage(dest_path)
        source_size = sum(f.stat().st_size for f in install_path.glob("**/*") if f.is_file())

        # Calculate from bytes to gigabytes
        free_space_dest_drive = round(free_space / 1000 ** 3, 2)
        source_size = round(source_size / 1000 ** 3, 2)
        self.aval_space_label.setText(self.tr("Available space: {}GB".format(free_space_dest_drive)))
        self.req_space_label.setText(self.tr("Required space: {}GB").format(source_size))

        if not os.access(dir_selected, os.W_OK) or not os.access(self.install_path, os.W_OK):
            return helper_func("No write permission on destination path/current install path.")

        if install_path == dest_path or install_path == dest_path_with_suffix:
            return helper_func("Same directory or parent directory selected.")

        if str(install_path) in str(dest_path):
            return helper_func("You can't select a directory that is inside the current install path.")

        if str(dest_path_with_suffix) in str(install_path):
            return helper_func("You can't select a directory which contains the game installation.")

        for game in self.core.get_installed_list():
            if game.install_path in dir_selected:
                return helper_func("Game installations cannot be nested due to unintended sideeffects.")

        is_existing_dir = is_game_dir(install_path, dest_path_with_suffix)

        for i in dest_path.iterdir():
            if install_path.stem in i.stem:
                if dest_path_with_suffix.is_dir():
                    if not is_existing_dir:
                        self.warn_overwriting.setHidden(False)
                elif dest_path_with_suffix.is_file():
                    self.warn_overwriting.setHidden(False)

        if free_space_dest_drive <= source_size and not is_existing_dir:
            return helper_func("Not enough space available on drive.")

        # Fallback
        self.move_game.setEnabled(True)
        return True, dir_selected, str()

    def update_game(self, app_name):
        igame = self.core.get_installed_game(app_name, skip_sync=True)
        if igame is None:
            return
        self.install_path = igame.install_path
        # FIXME: Make edit_func lighter instead of blocking signals
        self.move_path_edit.line_edit.blockSignals(True)
        self.move_path_edit.setText(igame.install_path)
        # FIXME: Make edit_func lighter instead of blocking signals
        self.move_path_edit.line_edit.blockSignals(False)
        self.warn_overwriting.setText(
            self.tr("Moving here will overwrite the dir/file {}/").format(Path(self.install_path).stem)
        )


# noinspection PyUnresolvedReferences
class CopyGameInstallation(QRunnable):
    class Signals(QObject):
        progress = pyqtSignal(int)
        finished = pyqtSignal(str)
        no_space_left = pyqtSignal()

    def __init__(
            self,
            install_path: str,
            dest_path: Path,
            is_existing_dir: bool,
            igame: InstalledGame,
    ):
        super(CopyGameInstallation, self).__init__()
        self.signals = CopyGameInstallation.Signals()
        self.install_path = install_path
        self.dest_path = dest_path
        self.source_size = 0
        self.dest_size = 0
        self.is_existing_dir = is_existing_dir
        self.core = LegendaryCoreSingleton()
        self.igame = igame
        self.file_list = None
        self.total: int = 0

    def run(self):
        root_directory = Path(self.install_path)
        self.source_size = sum(f.stat().st_size for f in root_directory.glob("**/*") if f.is_file())

        # if game dir is not existing, just copying:
        if not self.is_existing_dir:
            shutil.copytree(
                self.install_path,
                self.dest_path,
                copy_function=self.copy_each_file_with_progress,
                dirs_exist_ok=True,
            )
        else:
            manifest_data, _ = self.core.get_installed_manifest(self.igame.app_name)
            manifest = self.core.load_manifest(manifest_data)
            files = sorted(
                manifest.file_manifest_list.elements,
                key=lambda a: a.filename.lower(),
            )
            self.file_list = [(f.filename, f.sha_hash.hex()) for f in files]
            self.total = len(self.file_list)

            # recreate dir structure
            shutil.copytree(
                self.install_path,
                self.dest_path,
                copy_function=self.copy_dir_structure,
                dirs_exist_ok=True,
            )

            for i, (result, relative_path, _, _) in enumerate(
                    validate_files(str(self.dest_path), self.file_list)
            ):
                dst_path = f"{self.dest_path}/{relative_path}"
                src_path = f"{self.install_path}/{relative_path}"
                if Path(src_path).is_file():
                    if result == VerifyResult.HASH_MISMATCH:
                        try:
                            shutil.copy(src_path, dst_path)
                        except IOError:
                            self.signals.no_space_left.emit()
                            return
                    elif result == VerifyResult.FILE_MISSING:
                        try:
                            shutil.copy(src_path, dst_path)
                        except (IOError, OSError):
                            self.signals.no_space_left.emit()
                            return
                    elif result == VerifyResult.OTHER_ERROR:
                        logger.warning(f"Copying file {src_path} to {dst_path} failed")
                    self.signals.progress.emit(int(i * 10 / self.total * 10))
                else:
                    logger.warning(
                        f"Source dir does not have file {src_path}. File will be missing in the destination "
                        f"dir. "
                    )

        shutil.rmtree(self.install_path)
        self.signals.finished.emit(str(self.dest_path))

    def copy_each_file_with_progress(self, src, dst):
        shutil.copy(src, dst)
        self.dest_size += Path(src).stat().st_size
        self.signals.progress.emit(int(self.dest_size * 10 / self.source_size * 10))

    # This method is a copy_func, and only copies the src if it's a dir.
    # Thus, it can be used to re-create the dir strucute.
    @staticmethod
    def copy_dir_structure(src, dst):
        if os.path.isdir(dst):
            dst = os.path.join(dst, os.path.basename(src))
        if os.path.isdir(src):
            shutil.copyfile(src, dst)
            shutil.copystat(src, dst)
        return dst


def is_game_dir(install_path: Path, dest_path: Path):
    # This iterates over the destination dir, then iterates over the current install dir and if the file names
    # matches, we have an exisiting dir
    if dest_path.is_dir():
        for file in dest_path.iterdir():
            for install_file in install_path.iterdir():
                if file.name == install_file.name:
                    return True
    return False



import json
from dataclasses import dataclass
from logging import getLogger
from typing import Callable
from typing import Union

from PyQt5.QtCore import QObject, pyqtSignal, QUrl, QJsonParseError, QJsonDocument
from PyQt5.QtNetwork import QNetworkAccessManager, QNetworkRequest, QNetworkReply

logger = getLogger("QtRequests")


class QtRequestManager(QObject):
    data_ready = pyqtSignal(object)
    request = None
    request_active = None

    def __init__(self, type: str = "json", authorization_token: str = None):
        super(QtRequestManager, self).__init__()
        self.manager = QNetworkAccessManager()
        self.type = type
        self.authorization_token = authorization_token
        self.request_queue = []

    def post(self, url: str, payload: dict, handle_func):
        if not self.request_active:
            request = QNetworkRequest(QUrl(url))
            request.setHeader(QNetworkRequest.ContentTypeHeader, "application/json")
            self.request_active = RequestQueueItem(handle_func=handle_func)
            payload = json.dumps(payload).encode("utf-8")

            request.setHeader(
                QNetworkRequest.UserAgentHeader,
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.159 Safari/537.36",
            )

            if self.authorization_token is not None:
                request.setRawHeader(
                    b"Authorization", self.authorization_token.encode()
                )

            self.request = self.manager.post(request, payload)
            self.request.finished.connect(self.prepare_data)

        else:
            self.request_queue.append(
                RequestQueueItem(
                    method="post", url=url, payload=payload, handle_func=handle_func
                )
            )

    def get(self, url: str, handle_func: Callable[[Union[dict, bytes]], None]):
        if not self.request_active:
            request = QNetworkRequest(QUrl(url))
            request.setHeader(
                QNetworkRequest.UserAgentHeader,
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.159 Safari/537.36",
            )

            self.request_active = RequestQueueItem(handle_func=handle_func)
            self.request = self.manager.get(request)
            self.request.finished.connect(self.prepare_data)
        else:
            self.request_queue.append(
                RequestQueueItem(method="get", url=url, handle_func=handle_func)
            )

    def prepare_data(self):
        # self.request_active = False
        data = {} if self.type == "json" else b""
        if self.request:
            try:
                if self.request.error() == QNetworkReply.NoError:
                    if self.type == "json":
                        error = QJsonParseError()
                        json_data = QJsonDocument.fromJson(
                            self.request.readAll().data(), error
                        )
                        if QJsonParseError.NoError == error.error:
                            data = json.loads(json_data.toJson().data().decode())
                        else:
                            logger.error(error.errorString())
                    else:
                        data = self.request.readAll().data()

            except RuntimeError as e:
                logger.error(str(e))
        self.request_active.handle_func(data)
        self.request.deleteLater()
        self.request_active = None

        if self.request_queue:
            if self.request_queue[0].method == "post":
                self.post(
                    self.request_queue[0].url,
                    self.request_queue[0].payload,
                    self.request_queue[0].handle_func,
                )
            else:
                self.get(self.request_queue[0].url, self.request_queue[0].handle_func)
            self.request_queue.pop(0)


@dataclass
class RequestQueueItem:
    method: str = None
    url: str = None
    handle_func: Callable[[Union[dict, bytes]], None] = None
    payload: dict = None



import json
import logging
import sys
import time
import traceback
from argparse import Namespace
from logging import getLogger
from typing import Union, Optional

from PyQt5.QtCore import QObject, QProcess, pyqtSignal, QUrl, QRunnable, QThreadPool, QSettings, Qt
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtNetwork import QLocalServer, QLocalSocket
from PyQt5.QtWidgets import QApplication

from rare.lgndr.core import LegendaryCore
from rare.widgets.rare_app import RareApp
from .console import Console
from .lgd_helper import get_launch_args, InitArgs, get_configured_process, LaunchArgs, GameArgsError
from .message_models import ErrorModel, Actions, FinishedModel, BaseModel, StateChangedModel

logger = logging.getLogger("RareLauncher")


class PreLaunchThread(QRunnable):
    class Signals(QObject):
        ready_to_launch = pyqtSignal(LaunchArgs)
        started_pre_launch_command = pyqtSignal()
        pre_launch_command_finished = pyqtSignal(int)  # exit_code
        error_occurred = pyqtSignal(str)

    def __init__(self, core: LegendaryCore, args: InitArgs):
        super(PreLaunchThread, self).__init__()
        self.core = core
        self.app_name = args.app_name
        self.signals = self.Signals()
        self.args = args

    def run(self) -> None:
        args = self.prepare_launch(self.args)
        if not args:
            return
        self.signals.ready_to_launch.emit(args)

    def prepare_launch(self, args: InitArgs) -> Union[LaunchArgs, None]:
        try:
            args = get_launch_args(self.core, args)
        except Exception as e:
            self.signals.error_occurred.emit(str(e))
            return None
        if not args:
            return None

        if args.pre_launch_command:
            proc = get_configured_process()
            proc.setProcessEnvironment(args.env)
            self.signals.started_pre_launch_command.emit()
            proc.start(args.pre_launch_command[0], args.pre_launch_command[1:])
            if args.pre_launch_wait:
                proc.waitForFinished(-1)
        return args


class GameProcessApp(RareApp):
    game_process: QProcess
    server: QLocalServer
    socket: Optional[QLocalSocket] = None
    exit_app = pyqtSignal()
    console: Optional[Console] = None
    success: bool = True

    def __init__(self, args: Namespace):
        log_file = f"Rare_Launcher_{args.app_name}" + "_{0}.log"
        super(GameProcessApp, self).__init__(args, log_file)
        self.game_process = QProcess()
        self.app_name = args.app_name
        self.logger = getLogger(self.app_name)
        self.core = LegendaryCore()

        lang = self.settings.value("language", self.core.language_code, type=str)
        self.load_translator(lang)

        if QSettings().value("show_console", False, bool):
            self.console = Console()
            self.console.show()

        self.server = QLocalServer()
        ret = self.server.listen(f"rare_{self.app_name}")
        if not ret:
            self.logger.error(self.server.errorString())
            self.logger.info("Server is running")
            self.server.close()
            self.success = False
            return
        self.server.newConnection.connect(self.new_server_connection)

        self.game_process.finished.connect(self.game_finished)
        self.game_process.errorOccurred.connect(
            lambda err: self.error_occurred(self.game_process.errorString()))
        if self.console:
            self.game_process.readyReadStandardOutput.connect(
                lambda: self.console.log(
                    self.game_process.readAllStandardOutput().data().decode("utf-8", "ignore")
                )
            )
            self.game_process.readyReadStandardError.connect(
                lambda: self.console.error(
                    self.game_process.readAllStandardError().data().decode("utf-8", "ignore")
                )
            )
            self.console.term.connect(lambda: self.game_process.terminate())
            self.console.kill.connect(lambda: self.game_process.kill())

        self.start_time = time.time()

    def new_server_connection(self):
        if self.socket is not None:
            try:
                self.socket.disconnectFromServer()
            except RuntimeError:
                pass
        self.logger.info("New connection")
        self.socket = self.server.nextPendingConnection()
        self.socket.disconnected.connect(self.socket_disconnected)
        self.socket.flush()

    def socket_disconnected(self):
        self.logger.info("Server disconnected")
        self.socket.deleteLater()
        self.socket = None

    def send_message(self, message: BaseModel):
        if self.socket:
            self.socket.write(json.dumps(message.__dict__).encode("utf-8"))
            self.socket.flush()
        else:
            self.logger.error("Can't send message")

    def game_finished(self, exit_code):
        self.logger.info("game finished")
        if self.console:
            self.console.on_process_exit(self.core.get_game(self.app_name).app_title, exit_code)
        self.send_message(
            FinishedModel(
                action=Actions.finished,
                app_name=self.app_name,
                exit_code=exit_code,
                playtime=int(time.time() - self.start_time)
            )

        )
        self.stop()

    def launch_game(self, args: LaunchArgs):
        # should never happen
        if not args:
            self.stop()
            return
        if self.console:
            self.console.set_env(args.env)
        self.start_time = time.time()

        if args.is_origin_game:
            QDesktopServices.openUrl(QUrl(args.executable))
            self.stop()  # stop because it is no subprocess
            return

        if args.cwd:
            self.game_process.setWorkingDirectory(args.cwd)
        self.game_process.setProcessEnvironment(args.env)
        # send start message after process started
        self.game_process.started.connect(lambda: self.send_message(
            StateChangedModel(
                action=Actions.state_update, app_name=self.app_name,
                new_state=StateChangedModel.States.started
            )
        ))
        self.game_process.start(args.executable, args.args)

    def error_occurred(self, error_str: str):
        self.logger.warning(error_str)
        if self.console:
            self.console.on_process_exit(self.core.get_game(self.app_name).app_title, error_str)
        self.send_message(ErrorModel(
            error_string=error_str, app_name=self.app_name,
            action=Actions.error)
        )
        self.stop()

    def start(self, args: InitArgs):
        if not args.offline:
            try:
                if not self.core.login():
                    raise ValueError("You are not logged in")
            except ValueError:
                # automatically launch offline if available
                self.logger.error("Not logged in. Try to launch game offline")
                args.offline = True

        worker = PreLaunchThread(self.core, args)
        worker.signals.ready_to_launch.connect(self.launch_game)
        worker.signals.error_occurred.connect(self.error_occurred)
        # worker.signals.started_pre_launch_command(None)

        QThreadPool.globalInstance().start(worker)

    def stop(self):
        self.logger.info("Stopping server")
        try:
            self.server.close()
            self.server.deleteLater()
        except RuntimeError:
            pass
        self.processEvents()
        if not self.console:
            self.exit()


def start_game(args: Namespace):
    args = InitArgs.from_argparse(args)

    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)

    app = GameProcessApp(args)
    app.setQuitOnLastWindowClosed(True)

    def excepthook(exc_type, exc_value, exc_tb):
        tb = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
        app.logger.fatal(tb)
        try:
            app.send_message(ErrorModel(
                app_name=args.app_name,
                action=Actions.error,
                error_string=tb
            ))
        except RuntimeError:
            pass
        app.stop()

    sys.excepthook = excepthook
    if not app.success:
        return
    app.start(args)
    # app.exit_app.connect(lambda: app.exit(0))

    sys.exit(app.exec_())



import json
import os
from dataclasses import dataclass
from datetime import datetime
from logging import getLogger
from typing import Dict

from rare.utils.paths import data_dir

logger = getLogger("GameMeta")


@dataclass
class GameMeta:
    app_name: str
    last_played: datetime = None

    @classmethod
    def from_json(cls, data):
        return cls(
            app_name=data.get("app_name", ""),
            last_played=datetime.strptime(data.get("last_played", "None"), '%Y-%m-%dT%H:%M:%S.%f')
        )

    def __dict__(self):
        return dict(
            app_name=self.app_name,
            last_played=self.last_played.strftime("%Y-%m-%dT%H:%M:%S.%f")
        )


class RareGameMeta:
    _meta: Dict[str, GameMeta] = {}

    def __init__(self):
        meta_data = {}
        if os.path.exists(p := os.path.join(data_dir(), "game_meta.json")):
            try:
                meta_data = json.load(open(p))
            except json.JSONDecodeError:
                logger.warning("Game meta json file corrupt")
        else:
            with open(p, "w") as file:
                file.write("{}")

        for app_name, data in meta_data.items():
            self._meta[app_name] = GameMeta.from_json(data)

    def get_games(self):
        return list(self._meta.values())

    def get_game(self, app_name):
        return self._meta.get(app_name, GameMeta(app_name))

    def set_game(self, app_name: str, game: GameMeta):
        self._meta[app_name] = game
        self.save_file()

    def save_file(self):
        json.dump(
            {app_name: data.__dict__() for app_name, data in self._meta.items()},
            open(os.path.join(data_dir(), "game_meta.json"), "w"),
            indent=4
        )



from PyQt5.QtCore import Qt
from PyQt5.QtGui import QKeyEvent

from rare.components.tabs.games.game_utils import GameUtils
from rare.shared import LegendaryCoreSingleton, GlobalSignalsSingleton
from rare.utils.extra_widgets import SideTabWidget
from .game_dlc import GameDlc
from .game_info import GameInfo
from .game_settings import GameSettings


class GameInfoTabs(SideTabWidget):
    def __init__(self, dlcs: dict, game_utils: GameUtils, parent=None):
        super(GameInfoTabs, self).__init__(show_back=True, parent=parent)
        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()

        self.info = GameInfo(self, game_utils)
        self.addTab(self.info, self.tr("Information"))

        self.settings = GameSettings(self)
        self.addTab(self.settings, self.tr("Settings"))

        self.dlc_list = dlcs
        self.dlc = GameDlc(self.dlc_list, game_utils, self)
        self.addTab(self.dlc, self.tr("Downloadable Content"))

        self.tabBar().setCurrentIndex(1)

    def update_game(self, app_name: str):
        self.setCurrentIndex(1)
        self.info.update_game(app_name)
        self.settings.load_settings(app_name)

        # DLC Tab: Disable if no dlcs available
        if (
                len(self.dlc_list.get(self.core.get_game(app_name).catalog_item_id, []))
                == 0
        ):
            self.setTabEnabled(3, False)
        else:
            self.setTabEnabled(3, True)
            self.dlc.update_dlcs(app_name)

    def keyPressEvent(self, e: QKeyEvent):
        if e.key() == Qt.Key_Escape:
            self.parent().layout().setCurrentIndex(0)



import logging
# import warnings


class LgndrException(RuntimeError):
    def __init__(self, message="Error in Legendary"):
        self.message = message
        super(LgndrException, self).__init__(self.message)


class LgndrWarning(RuntimeWarning):
    def __init__(self, message="Warning in Legendary"):
        self.message = message
        super(LgndrWarning, self).__init__(self.message)


class LgndrCLILogHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        # lk: FATAL is the same as CRITICAL
        if record.levelno == logging.ERROR or record.levelno == logging.CRITICAL:
            raise LgndrException(record.getMessage())
        # if record.levelno < logging.ERROR or record.levelno == logging.WARNING:
        #     warnings.warn(record.getMessage())


class LgndrCoreLogHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        # lk: FATAL is the same as CRITICAL
        if record.levelno == logging.CRITICAL:
            raise LgndrException(record.getMessage())
        # if record.levelno < logging.CRITICAL:
        #     warnings.warn(record.getMessage())



from multiprocessing import Queue

# On Windows the monkeypatching of `run_real` below doesn't work like on Linux
# This has the side effect of emitting the UIUpdate in DownloadThread complaining with a TypeError
# So import `legendary.core` and monkeypatch its imported DLManager
import legendary.core
from legendary.core import LegendaryCore as LegendaryCoreReal
from legendary.models.downloading import AnalysisResult
from legendary.models.game import Game, InstalledGame
from legendary.models.manifest import ManifestMeta

from rare.lgndr.downloader.mp.manager import DLManager
from rare.lgndr.glue.exception import LgndrException, LgndrCoreLogHandler

legendary.core.DLManager = DLManager


# fmt: off
class LegendaryCore(LegendaryCoreReal):

    def __init__(self, override_config=None, timeout=10.0):
        super(LegendaryCore, self).__init__(override_config=override_config, timeout=timeout)
        self.handler = LgndrCoreLogHandler()
        self.log.addHandler(self.handler)

    # skip_sync defaults to false but since Rare is persistent, skip by default
    # def get_installed_game(self, app_name, skip_sync=True) -> InstalledGame:
    #     return super(LegendaryCore, self).get_installed_game(app_name, skip_sync)

    def prepare_download(self, game: Game, base_game: Game = None, base_path: str = '',
                         status_q: Queue = None, max_shm: int = 0, max_workers: int = 0,
                         force: bool = False, disable_patching: bool = False,
                         game_folder: str = '', override_manifest: str = '',
                         override_old_manifest: str = '', override_base_url: str = '',
                         platform: str = 'Windows', file_prefix_filter: list = None,
                         file_exclude_filter: list = None, file_install_tag: list = None,
                         dl_optimizations: bool = False, dl_timeout: int = 10,
                         repair: bool = False, repair_use_latest: bool = False,
                         disable_delta: bool = False, override_delta_manifest: str = '',
                         egl_guid: str = '', preferred_cdn: str = None,
                         disable_https: bool = False) -> (DLManager, AnalysisResult, ManifestMeta):
        dlm, analysis, igame = super(LegendaryCore, self).prepare_download(
            game=game, base_game=base_game, base_path=base_path,
            status_q=status_q, max_shm=max_shm, max_workers=max_workers,
            force=force, disable_patching=disable_patching,
            game_folder=game_folder, override_manifest=override_manifest,
            override_old_manifest=override_old_manifest, override_base_url=override_base_url,
            platform=platform, file_prefix_filter=file_prefix_filter,
            file_exclude_filter=file_exclude_filter, file_install_tag=file_install_tag,
            dl_optimizations=dl_optimizations, dl_timeout=dl_timeout,
            repair=repair, repair_use_latest=repair_use_latest,
            disable_delta=disable_delta, override_delta_manifest=override_delta_manifest,
            egl_guid=egl_guid, preferred_cdn=preferred_cdn,
            disable_https=disable_https
        )
        # lk: monkeypatch run_real (the method that emits the stats) into DLManager

        # pylint: disable=E1111
        dlm.run_real = DLManager.run_real.__get__(dlm, DLManager)
        # lk: set the queue for reporting statistics back the UI
        dlm.status_queue = Queue()
        # lk: set the queue to send control signals to the DLManager
        # lk: this doesn't exist in the original class, but it is monkeypatched in
        dlm.signals_queue = Queue()
        return dlm, analysis, igame

    def uninstall_game(self, installed_game: InstalledGame, delete_files=True, delete_root_directory=False):
        try:
            super(LegendaryCore, self).uninstall_game(installed_game, delete_files, delete_root_directory)
        except Exception as e:
            raise e
        finally:
            pass

    def egl_import(self, app_name):
        try:
            super(LegendaryCore, self).egl_import(app_name)
        except LgndrException as ret:
            raise ret
        finally:
            pass

    def egl_export(self, app_name):
        try:
            super(LegendaryCore, self).egl_export(app_name)
        except LgndrException as ret:
            raise ret
        finally:
            pass

    def prepare_overlay_install(self, path=None):
        dlm, analysis_result, igame = super(LegendaryCore, self).prepare_overlay_install(path)
        # lk: monkeypatch status_q (the queue for download stats)
        # pylint: disable=E1111
        dlm.run_real = DLManager.run_real.__get__(dlm, DLManager)
        # lk: set the queue for reporting statistics back the UI
        dlm.status_queue = Queue()
        # lk: set the queue to send control signals to the DLManager
        # lk: this doesn't exist in the original class, but it is monkeypatched in
        dlm.signals_queue = Queue()
        return dlm, analysis_result, igame

# fmt: on



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/downloads/downloads_tab.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_DownloadsTab(object):
    def setupUi(self, DownloadsTab):
        DownloadsTab.setObjectName("DownloadsTab")
        DownloadsTab.resize(677, 195)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(DownloadsTab.sizePolicy().hasHeightForWidth())
        DownloadsTab.setSizePolicy(sizePolicy)
        DownloadsTab.setWindowTitle("Form")
        self.downloads_tab_layout = QtWidgets.QVBoxLayout(DownloadsTab)
        self.downloads_tab_layout.setObjectName("downloads_tab_layout")
        self.download_layout = QtWidgets.QGridLayout()
        self.download_layout.setObjectName("download_layout")
        self.dl_speed_label = QtWidgets.QLabel(DownloadsTab)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.dl_speed_label.setFont(font)
        self.dl_speed_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.dl_speed_label.setObjectName("dl_speed_label")
        self.download_layout.addWidget(self.dl_speed_label, 1, 2, 1, 1)
        self.dl_speed = QtWidgets.QLabel(DownloadsTab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.dl_speed.sizePolicy().hasHeightForWidth())
        self.dl_speed.setSizePolicy(sizePolicy)
        self.dl_speed.setText("dl_speed")
        self.dl_speed.setObjectName("dl_speed")
        self.download_layout.addWidget(self.dl_speed, 1, 3, 1, 1)
        self.cache_used = QtWidgets.QLabel(DownloadsTab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cache_used.sizePolicy().hasHeightForWidth())
        self.cache_used.setSizePolicy(sizePolicy)
        self.cache_used.setText("cache_used")
        self.cache_used.setObjectName("cache_used")
        self.download_layout.addWidget(self.cache_used, 2, 1, 1, 1)
        self.time_left = QtWidgets.QLabel(DownloadsTab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.time_left.sizePolicy().hasHeightForWidth())
        self.time_left.setSizePolicy(sizePolicy)
        self.time_left.setText("time_left")
        self.time_left.setObjectName("time_left")
        self.download_layout.addWidget(self.time_left, 1, 1, 1, 1)
        self.dl_name = QtWidgets.QLabel(DownloadsTab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.dl_name.sizePolicy().hasHeightForWidth())
        self.dl_name.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.dl_name.setFont(font)
        self.dl_name.setText("dl_name")
        self.dl_name.setObjectName("dl_name")
        self.download_layout.addWidget(self.dl_name, 0, 1, 1, 3)
        self.dl_name_label = QtWidgets.QLabel(DownloadsTab)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.dl_name_label.setFont(font)
        self.dl_name_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.dl_name_label.setObjectName("dl_name_label")
        self.download_layout.addWidget(self.dl_name_label, 0, 0, 1, 1)
        self.time_left_label = QtWidgets.QLabel(DownloadsTab)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.time_left_label.setFont(font)
        self.time_left_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.time_left_label.setObjectName("time_left_label")
        self.download_layout.addWidget(self.time_left_label, 1, 0, 1, 1)
        self.downloaded_label = QtWidgets.QLabel(DownloadsTab)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.downloaded_label.setFont(font)
        self.downloaded_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.downloaded_label.setObjectName("downloaded_label")
        self.download_layout.addWidget(self.downloaded_label, 2, 2, 1, 1)
        self.downloaded = QtWidgets.QLabel(DownloadsTab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.downloaded.sizePolicy().hasHeightForWidth())
        self.downloaded.setSizePolicy(sizePolicy)
        self.downloaded.setText("downloaded")
        self.downloaded.setObjectName("downloaded")
        self.download_layout.addWidget(self.downloaded, 2, 3, 1, 1)
        self.cache_used_label = QtWidgets.QLabel(DownloadsTab)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.cache_used_label.setFont(font)
        self.cache_used_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.cache_used_label.setObjectName("cache_used_label")
        self.download_layout.addWidget(self.cache_used_label, 2, 0, 1, 1)
        self.download_layout.setColumnStretch(1, 1)
        self.download_layout.setColumnStretch(3, 1)
        self.downloads_tab_layout.addLayout(self.download_layout)
        self.progress_layout = QtWidgets.QHBoxLayout()
        self.progress_layout.setObjectName("progress_layout")
        self.progress_bar = QtWidgets.QProgressBar(DownloadsTab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progress_bar.sizePolicy().hasHeightForWidth())
        self.progress_bar.setSizePolicy(sizePolicy)
        self.progress_bar.setProperty("value", 0)
        self.progress_bar.setObjectName("progress_bar")
        self.progress_layout.addWidget(self.progress_bar)
        self.kill_button = QtWidgets.QPushButton(DownloadsTab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.kill_button.sizePolicy().hasHeightForWidth())
        self.kill_button.setSizePolicy(sizePolicy)
        self.kill_button.setObjectName("kill_button")
        self.progress_layout.addWidget(self.kill_button)
        self.downloads_tab_layout.addLayout(self.progress_layout)
        self.queue_scroll = QtWidgets.QScrollArea(DownloadsTab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.queue_scroll.sizePolicy().hasHeightForWidth())
        self.queue_scroll.setSizePolicy(sizePolicy)
        self.queue_scroll.setFocusPolicy(QtCore.Qt.WheelFocus)
        self.queue_scroll.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.queue_scroll.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.queue_scroll.setWidgetResizable(True)
        self.queue_scroll.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.queue_scroll.setObjectName("queue_scroll")
        self.queue_scroll_contents = QtWidgets.QWidget()
        self.queue_scroll_contents.setGeometry(QtCore.QRect(0, 0, 665, 16))
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.queue_scroll_contents.sizePolicy().hasHeightForWidth())
        self.queue_scroll_contents.setSizePolicy(sizePolicy)
        self.queue_scroll_contents.setObjectName("queue_scroll_contents")
        self.queue_scroll_contents_layout = QtWidgets.QVBoxLayout(self.queue_scroll_contents)
        self.queue_scroll_contents_layout.setContentsMargins(0, 0, 9, 0)
        self.queue_scroll_contents_layout.setObjectName("queue_scroll_contents_layout")
        self.queue_scroll.setWidget(self.queue_scroll_contents)
        self.downloads_tab_layout.addWidget(self.queue_scroll)

        self.retranslateUi(DownloadsTab)
        QtCore.QMetaObject.connectSlotsByName(DownloadsTab)

    def retranslateUi(self, DownloadsTab):
        _translate = QtCore.QCoreApplication.translate
        self.dl_speed_label.setText(_translate("DownloadsTab", "Download speed:"))
        self.dl_name_label.setText(_translate("DownloadsTab", "Current download:"))
        self.time_left_label.setText(_translate("DownloadsTab", "Time left:"))
        self.downloaded_label.setText(_translate("DownloadsTab", "Downloaded:"))
        self.cache_used_label.setText(_translate("DownloadsTab", "Cache size:"))
        self.kill_button.setText(_translate("DownloadsTab", "Stop Download"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    DownloadsTab = QtWidgets.QWidget()
    ui = Ui_DownloadsTab()
    ui.setupUi(DownloadsTab)
    DownloadsTab.show()
    sys.exit(app.exec_())



from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QFrame
from legendary.models.game import Game

from rare.shared import LegendaryCoreSingleton, ImageManagerSingleton
from rare.shared.image_manager import ImageSize
from rare.widgets.elide_label import ElideLabel
from .library_widget import LibraryWidget


class InstallingGameWidget(QFrame):
    game: Game = None

    def __init__(self):
        super(InstallingGameWidget, self).__init__()
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setFixedWidth(ImageSize.Display.size.width())
        self.setObjectName(type(self).__name__)

        self.core = LegendaryCoreSingleton()
        self.image_manager = ImageManagerSingleton()

        self.pixmap = QPixmap()
        self.image = LibraryWidget(parent=self)
        self.image.setFixedSize(ImageSize.Display)
        layout.addWidget(self.image)

        miniwidget = QWidget(self)
        miniwidget.setFixedWidth(ImageSize.Display.size.width())
        minilayout = QHBoxLayout()
        minilayout.setContentsMargins(0, 0, 0, 0)
        minilayout.setSpacing(0)
        miniwidget.setLayout(minilayout)

        self.title_label = ElideLabel(f"<h4>Error</h4>", parent=miniwidget)
        self.title_label.setAlignment(Qt.AlignTop)
        self.title_label.setObjectName("game_widget")
        minilayout.addWidget(self.title_label, stretch=2)

        minilayout.setAlignment(Qt.AlignTop)
        layout.addWidget(miniwidget)

        self.setLayout(layout)

    def set_game(self, app_name):
        self.game = self.core.get_game(app_name, False)
        if (not self.game) or self.game.is_dlc:
            # Don't show for EOS Overlay or DLCs
            self.game = None
            self.setVisible(False)
            return
        self.setVisible(True)
        self.title_label.setText(f"<h4>{self.game.app_title}</h4>")
        self.image.hideProgress(True)
        self.image.showProgress(
            self.image_manager.get_pixmap(app_name, color=True),
            self.image_manager.get_pixmap(app_name, color=False),
        )

    def set_status(self, s: int):
        if not self.game:
            # Don't show for EOS Overlay or DLCs
            return
        self.image.updateProgress(s)



import os
import platform
import shlex
import subprocess
import sys
from logging import getLogger
from typing import List, Union

import qtawesome
import requests
from PyQt5.QtCore import (
    pyqtSignal,
    pyqtSlot,
    QObject,
    QRunnable,
    QSettings,
    QStandardPaths,
    QFile,
    QDir,
)
from PyQt5.QtGui import QPalette, QColor, QImage
from PyQt5.QtWidgets import qApp, QStyleFactory
from legendary.models.game import Game
from requests.exceptions import HTTPError

from .models import PathSpec

# Windows

if platform.system() == "Windows":
    # noinspection PyUnresolvedReferences
    from win32com.client import Dispatch  # pylint: disable=E0401

from rare.shared import LegendaryCoreSingleton, ApiResultsSingleton
from rare.utils.paths import image_dir, resources_path

# Mac not supported

from legendary.core import LegendaryCore

logger = getLogger("Utils")
settings = QSettings("Rare", "Rare")

color_role_map = {
    0: "WindowText",
    1: "Button",
    2: "Light",
    3: "Midlight",
    4: "Dark",
    5: "Mid",
    6: "Text",
    7: "BrightText",
    8: "ButtonText",
    9: "Base",
    10: "Window",
    11: "Shadow",
    12: "Highlight",
    13: "HighlightedText",
    14: "Link",
    15: "LinkVisited",
    16: "AlternateBase",
    # 17: "NoRole",
    18: "ToolTipBase",
    19: "ToolTipText",
    20: "PlaceholderText",
    # 21: "NColorRoles",
}

color_group_map = {
    0: "Active",
    1: "Disabled",
    2: "Inactive",
}


def load_color_scheme(path: str) -> QPalette:
    palette = QPalette()
    scheme = QSettings(path, QSettings.IniFormat)
    try:
        scheme.beginGroup("ColorScheme")
        for g in color_group_map:
            scheme.beginGroup(color_group_map[g])
            group = QPalette.ColorGroup(g)
            for r in color_role_map:
                role = QPalette.ColorRole(r)
                color = scheme.value(color_role_map[r], None)
                if color is not None:
                    palette.setColor(group, role, QColor(color))
                else:
                    palette.setColor(group, role, palette.color(QPalette.Active, role))
            scheme.endGroup()
        scheme.endGroup()
    except:
        palette = None
    return palette


def set_color_pallete(color_scheme: str):
    if not color_scheme:
        qApp.setStyle(QStyleFactory.create(qApp.property("rareDefaultQtStyle")))
        qApp.setStyleSheet("")
        qApp.setPalette(qApp.style().standardPalette())
        return
    qApp.setStyle(QStyleFactory.create("Fusion"))
    custom_palette = load_color_scheme(f":/schemes/{color_scheme}")
    if custom_palette is not None:
        qApp.setPalette(custom_palette)
        icon_color = qApp.palette().color(QPalette.Foreground).name()
        qtawesome.set_defaults(color=icon_color)


def get_color_schemes() -> List[str]:
    colors = []
    for file in QDir(":/schemes"):
        colors.append(file)
    return colors


def set_style_sheet(style_sheet: str):
    if not style_sheet:
        qApp.setStyle(QStyleFactory.create(qApp.property("rareDefaultQtStyle")))
        qApp.setStyleSheet("")
        return
    qApp.setStyle(QStyleFactory.create("Fusion"))
    file = QFile(f":/stylesheets/{style_sheet}/stylesheet.qss")
    file.open(QFile.ReadOnly)
    stylesheet = file.readAll().data().decode("utf-8")

    qApp.setStyleSheet(stylesheet)
    icon_color = qApp.palette().color(QPalette.Text).name()
    qtawesome.set_defaults(color="#eeeeee")


def get_style_sheets() -> List[str]:
    styles = []
    for file in QDir(":/stylesheets/"):
        styles.append(file)
    return styles


def get_translations():
    langs = ["en"]
    for i in os.listdir(os.path.join(resources_path, "languages")):
        if i.endswith(".qm") and not i.startswith("qt_"):
            langs.append(i.split(".")[0])
    return langs


def get_latest_version():
    try:
        resp = requests.get(
            "https://api.github.com/repos/Dummerle/Rare/releases/latest"
        )
        tag = resp.json()["tag_name"]
        return tag
    except requests.exceptions.ConnectionError:
        return "0.0.0"


def get_size(b: Union[int, float]) -> str:
    for i in ["", "K", "M", "G", "T", "P", "E"]:
        if b < 1024:
            return f"{b:.2f}{i}B"
        b /= 1024


def get_rare_executable() -> List[str]:
    # lk: detect if nuitka
    if "__compiled__" in globals():
        executable = [sys.executable]
    elif platform.system() == "Linux" or platform.system() == "Darwin":
        if p := os.environ.get("APPIMAGE"):
            executable = [p]
        else:
            if sys.executable == os.path.abspath(sys.argv[0]):
                executable = [sys.executable]
            else:
                executable = [sys.executable, os.path.abspath(sys.argv[0])]
    elif platform.system() == "Windows":
        executable = [sys.executable]

        if sys.executable != os.path.abspath(sys.argv[0]):
            executable.append(os.path.abspath(sys.argv[0]))

        if executable[0].endswith("python.exe"):
            # be sure to start consoleless then
            executable[0] = executable[0].replace("python.exe", "pythonw.exe")
            if executable[1].endswith("rare"):
                executable[1] = executable[1] + ".exe"
    else:
        executable = [sys.executable]

    executable[0] = os.path.abspath(executable[0])
    return executable


def create_desktop_link(app_name=None, core: LegendaryCore = None, type_of_link="desktop",
                        for_rare: bool = False) -> bool:
    if not for_rare:
        igame = core.get_installed_game(app_name)

        icon = os.path.join(os.path.join(image_dir(), igame.app_name, "installed.png"))
        icon = icon.replace(".png", "")

    if platform.system() == "Linux":
        if type_of_link == "desktop":
            path = QStandardPaths.writableLocation(QStandardPaths.DesktopLocation)
        elif type_of_link == "start_menu":
            path = QStandardPaths.writableLocation(QStandardPaths.ApplicationsLocation)
        else:
            return False
        if not os.path.exists(path):
            return False
        executable = get_rare_executable()
        executable = shlex.join(executable)

        if for_rare:
            with open(os.path.join(path, "Rare.desktop"), "w") as desktop_file:
                desktop_file.write(
                    "[Desktop Entry]\n"
                    f"Name=Rare\n"
                    f"Type=Application\n"
                    f"Categories=Game;\n"
                    f"Icon={os.path.join(resources_path, 'images', 'Rare.png')}\n"
                    f"Exec={executable}\n"
                    "Terminal=false\n"
                    "StartupWMClass=Rare\n"
                )
        else:
            with open(os.path.join(path, f"{igame.title}.desktop"), "w") as desktop_file:
                desktop_file.write(
                    "[Desktop Entry]\n"
                    f"Name={igame.title}\n"
                    f"Type=Application\n"
                    f"Categories=Game;\n"
                    f"Icon={icon}.png\n"
                    f"Exec={executable} launch {app_name}\n"
                    "Terminal=false\n"
                    "StartupWMClass=Rare\n"
                )
            os.chmod(os.path.join(path, f"{igame.title}.desktop"), 0o755)

        return True

    elif platform.system() == "Windows":
        # Target of shortcut
        if type_of_link == "desktop":
            target_folder = QStandardPaths.writableLocation(QStandardPaths.DesktopLocation)
        elif type_of_link == "start_menu":
            target_folder = os.path.join(
                QStandardPaths.writableLocation(QStandardPaths.ApplicationsLocation),
                ".."
            )
        else:
            logger.warning("No valid type of link")
            return False
        if not os.path.exists(target_folder):
            return False

        if for_rare:
            linkName = "Rare.lnk"
        else:
            linkName = igame.title
            # TODO: this conversion is not applied everywhere (see base_installed_widget), should it?
            for c in r'<>?":|\/*':
                linkName = linkName.replace(c, "")

            linkName = f"{linkName.strip()}.lnk"

        # Path to location of link file
        pathLink = os.path.join(target_folder, linkName)

        # Add shortcut
        shell = Dispatch("WScript.Shell")
        shortcut = shell.CreateShortCut(pathLink)

        executable = get_rare_executable()
        arguments = []

        if len(executable) > 1:
            arguments.extend(executable[1:])
        executable = executable[0]

        if not for_rare:
            arguments.extend(["launch", app_name])

        shortcut.Targetpath = executable
        # Maybe there is a better solution, but windows does not accept single quotes (Windows is weird)
        shortcut.Arguments = shlex.join(arguments).replace("'", '"')
        if for_rare:
            shortcut.WorkingDirectory = QStandardPaths.writableLocation(QStandardPaths.HomeLocation)

        # Icon
        if for_rare:
            icon_location = os.path.join(resources_path, "images", "Rare.ico")
        else:
            if not os.path.exists(f"{icon}.ico"):
                img = QImage()
                img.load(f"{icon}.png")
                img.save(f"{icon}.ico")
                logger.info("Created ico file")
            icon_location = f"{icon}.ico"
        shortcut.IconLocation = os.path.abspath(icon_location)

        shortcut.save()
        return True

    # mac OS is based on Darwin
    elif platform.system() == "Darwin":
        return False


class WineResolverSignals(QObject):
    result_ready = pyqtSignal(str)


class WineResolver(QRunnable):
    def __init__(self, path: str, app_name: str):
        super(WineResolver, self).__init__()
        self.signals = WineResolverSignals()
        self.setAutoDelete(True)
        self.wine_env = os.environ.copy()
        core = LegendaryCoreSingleton()
        self.wine_env.update(core.get_app_environment(app_name))
        self.wine_env["WINEDLLOVERRIDES"] = "winemenubuilder=d;mscoree=d;mshtml=d;"
        self.wine_env["DISPLAY"] = ""

        self.wine_binary = core.lgd.config.get(
            app_name,
            "wine_executable",
            fallback=core.lgd.config.get("default", "wine_executable", fallback="wine"),
        )
        self.winepath_binary = os.path.join(
            os.path.dirname(self.wine_binary), "winepath"
        )
        self.path = PathSpec(core, app_name).cook(path)

    @pyqtSlot()
    def run(self):
        if "WINEPREFIX" not in self.wine_env or not os.path.exists(
                self.wine_env["WINEPREFIX"]
        ):
            # pylint: disable=E1136
            self.signals.result_ready[str].emit(str())
            return
        if not os.path.exists(self.wine_binary) or not os.path.exists(
                self.winepath_binary
        ):
            # pylint: disable=E1136
            self.signals.result_ready[str].emit(str())
            return
        path = self.path.strip().replace("/", "\\")
        # lk: if path does not exist form
        cmd = [self.wine_binary, "cmd", "/c", "echo", path]
        # lk: if path exists and needs a case sensitive interpretation form
        # cmd = [self.wine_binary, 'cmd', '/c', f'cd {path} & cd']
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=self.wine_env,
            shell=False,
            text=True,
        )
        out, err = proc.communicate()
        # Clean wine output
        out = out.strip().strip('"')
        proc = subprocess.Popen(
            [self.winepath_binary, "-u", out],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=self.wine_env,
            shell=False,
            text=True,
        )
        out, err = proc.communicate()
        real_path = os.path.realpath(out.strip())
        # pylint: disable=E1136
        self.signals.result_ready[str].emit(real_path)
        return


class CloudSignals(QObject):
    result_ready = pyqtSignal(list)  # List[SaveGameFile]


class CloudWorker(QRunnable):
    def __init__(self):
        super(CloudWorker, self).__init__()
        self.signals = CloudSignals()
        self.setAutoDelete(True)
        self.core = LegendaryCoreSingleton()

    def run(self) -> None:
        try:
            result = self.core.get_save_games()
        except HTTPError:
            result = None
        self.signals.result_ready.emit(result)


def get_raw_save_path(game: Game):
    if game.supports_cloud_saves:
        return (
            game.metadata.get("customAttributes", {})
                .get("CloudSaveFolder", {})
                .get("value")
        )


def get_default_platform(app_name):
    api_results = ApiResultsSingleton()
    if platform.system() != "Darwin" or app_name not in api_results.mac_games:
        return "Windows"
    else:
        return "Mac"


def icon(icn_str: str, fallback: str = None, **kwargs):
    try:
        return qtawesome.icon(icn_str, **kwargs)
    except Exception as e:
        if not fallback:
            logger.warning(f"{e} {icn_str}")
    if fallback:
        try:
            return qtawesome.icon(fallback, **kwargs)
        except Exception as e:
            logger.error(str(e))
    if kwargs.get("color"):
        kwargs["color"] = "red"
    return qtawesome.icon("ei.error", **kwargs)



from dataclasses import dataclass
from enum import IntEnum
from typing import Callable, List, Optional, Dict

from rare.lgndr.glue.monkeys import (
    LgndrIndirectStatus,
    GetBooleanChoiceProtocol,
    get_boolean_choice,
    verify_stdout,
    DLManagerSignals
)
from rare.lgndr.models.downloading import UIUpdate

"""
@dataclass(kw_only=True)
class LgndrCommonArgs:
    # keep this here for future reference
    # when we move to 3.10 we can use 'kw_only' to do dataclass inheritance
    app_name: str
    platform: str = "Windows"
    yes: bool = False
"""


@dataclass
class LgndrImportGameArgs:
    app_path: str
    app_name: str
    platform: str = "Windows"
    disable_check: bool = False
    skip_dlcs: bool = False
    with_dlcs: bool = False
    yes: bool = False
    # Rare: Extra arguments
    indirect_status: LgndrIndirectStatus = LgndrIndirectStatus()
    get_boolean_choice: GetBooleanChoiceProtocol = get_boolean_choice


@dataclass
class LgndrUninstallGameArgs:
    app_name: str
    keep_files: bool = False
    yes: bool = False
    # Rare: Extra arguments
    indirect_status: LgndrIndirectStatus = LgndrIndirectStatus()
    get_boolean_choice: GetBooleanChoiceProtocol = get_boolean_choice


@dataclass
class LgndrVerifyGameArgs:
    app_name: str
    # Rare: Extra arguments
    indirect_status: LgndrIndirectStatus = LgndrIndirectStatus()
    verify_stdout: Callable[[int, int, float, float], None] = verify_stdout


@dataclass
class LgndrInstallGameArgs:
    app_name: str
    base_path: str = ""
    shared_memory: int = 0
    max_workers: int = 0
    force: bool = False
    disable_patching: bool = False
    game_folder: str = ""
    override_manifest: str = ""
    override_old_manifest: str = ""
    override_base_url: str = ""
    platform: str = "Windows"
    file_prefix: List = None
    file_exclude_prefix: List = None
    install_tag: Optional[List[str]] = None
    order_opt: bool = False
    dl_timeout: int = 10
    repair_mode: bool = False
    repair_and_update: bool = False
    disable_delta: bool = False
    override_delta_manifest: str = ""
    egl_guid: str = ""
    preferred_cdn: str = None
    no_install: bool = False
    ignore_space: bool = False
    disable_sdl: bool = False
    reset_sdl: bool = False
    skip_sdl: bool = False
    disable_https: bool = False
    # FIXME: move to LgndrInstallGameRealArgs
    skip_dlcs: bool = False
    with_dlcs: bool = False
    # end of FIXME
    yes: bool = True
    # Rare: Extra arguments
    indirect_status: LgndrIndirectStatus = LgndrIndirectStatus()
    get_boolean_choice: GetBooleanChoiceProtocol = get_boolean_choice
    sdl_prompt: Callable[[str, str], List[str]] = lambda app_name, title: [""]
    verify_stdout: Callable[[int, int, float, float], None] = verify_stdout

    # def __post_init__(self):
    #     if self.sdl_prompt is None:
    #         self.sdl_prompt: Callable[[str, str], list] = \
    #             lambda app_name, title: self.install_tag if self.install_tag is not None else [""]


@dataclass
class LgndrInstallGameRealArgs:
    app_name: str
    platform: str = "Windows"
    repair_mode: bool = False
    repair_file: str = ""
    no_install: bool = False
    save_path: str = ""
    skip_dlcs: bool = False
    with_dlcs: bool = False
    dlm_debug: bool = False
    yes: bool = False
    # Rare: Extra arguments
    install_prereqs: bool = False
    indirect_status: LgndrIndirectStatus = LgndrIndirectStatus()
    ui_update: Callable[[UIUpdate], None] = lambda ui: None
    dlm_signals: DLManagerSignals = DLManagerSignals()


@dataclass
class LgndrInstallGameRealRet:
    class ReturnCode(IntEnum):
        ERROR = 1
        STOPPED = 2
        FINISHED = 3

    app_name: str
    ret_code: ReturnCode = ReturnCode.ERROR
    message: str = ""
    dlcs: Optional[List[Dict]] = None
    sync_saves: bool = False
    tip_url: str = ""
    shortcuts: bool = False



import os
import platform as pf
import sys
from typing import Tuple, List, Union, Optional

from PyQt5.QtCore import Qt, QObject, QRunnable, QThreadPool, pyqtSignal, pyqtSlot, QSettings
from PyQt5.QtGui import QCloseEvent, QKeyEvent
from PyQt5.QtWidgets import QDialog, QFileDialog, QCheckBox, QLayout, QWidget, QVBoxLayout, QApplication
from legendary.models.downloading import ConditionCheckResult
from legendary.models.game import Game
from legendary.utils.selective_dl import get_sdl_appname

from rare.lgndr.cli import LegendaryCLI
from rare.lgndr.core import LegendaryCore
from rare.lgndr.glue.arguments import LgndrInstallGameArgs
from rare.lgndr.glue.exception import LgndrException
from rare.lgndr.glue.monkeys import LgndrIndirectStatus
from rare.models.install import InstallDownloadModel, InstallQueueItemModel
from rare.shared import LegendaryCoreSingleton, ApiResultsSingleton, ArgumentsSingleton
from rare.ui.components.dialogs.install_dialog import Ui_InstallDialog
from rare.ui.components.dialogs.install_dialog_advanced import Ui_InstallDialogAdvanced
from rare.utils import config_helper
from rare.utils.extra_widgets import PathEdit
from rare.utils.misc import get_size
from rare.widgets.collapsible_widget import CollapsibleFrame


class InstallDialogAdvanced(CollapsibleFrame):
    def __init__(self, parent=None):
        widget = QWidget()
        title = widget.tr("Advanced Options")
        self.ui = Ui_InstallDialogAdvanced()
        self.ui.setupUi(widget)
        super(InstallDialogAdvanced, self).__init__(widget=widget, title=title, parent=parent)


class InstallDialog(QDialog):
    result_ready = pyqtSignal(InstallQueueItemModel)

    def __init__(self, dl_item: InstallQueueItemModel, update=False, repair=False, silent=False, parent=None):
        super(InstallDialog, self).__init__(parent)
        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint)
        self.ui = Ui_InstallDialog()
        self.ui.setupUi(self)

        self.core = LegendaryCoreSingleton()
        self.api_results = ApiResultsSingleton()
        self.dl_item = dl_item
        self.app_name = self.dl_item.options.app_name
        self.game = (
            self.core.get_game(self.app_name)
            if not self.dl_item.options.overlay
            else Game(app_name=self.app_name, app_title="Epic Overlay")
        )

        self.advanced = InstallDialogAdvanced(parent=self)
        self.ui.advanced_layout.addWidget(self.advanced)

        self.selectable = CollapsibleFrame(widget=None, title=self.tr("Optional downloads"), parent=self)
        self.ui.selectable_layout.addWidget(self.selectable)

        self.game_path = self.game.metadata.get("customAttributes", {}).get("FolderName", {}).get("value", "")

        self.update = update
        self.repair = repair
        self.silent = silent

        self.options_changed = False
        self.worker_running = False
        self.reject_close = True

        self.threadpool = QThreadPool(self)
        self.threadpool.setMaxThreadCount(1)

        header = self.tr("Update") if update else self.tr("Install")
        self.ui.install_dialog_label.setText(f'<h3>{header} "{self.game.app_title}"</h3>')
        self.setWindowTitle(f'{QApplication.instance().applicationName()} - {header} "{self.game.app_title}"')

        if not self.dl_item.options.base_path:
            self.dl_item.options.base_path = self.core.lgd.config.get(
                "Legendary", "install_dir", fallback=os.path.expanduser("~/legendary")
            )

        self.install_dir_edit = PathEdit(
            path=self.dl_item.options.base_path,
            file_type=QFileDialog.DirectoryOnly,
            edit_func=self.option_changed,
            parent=self,
        )
        self.ui.install_dir_layout.addWidget(self.install_dir_edit)

        if self.update:
            self.ui.install_dir_label.setEnabled(False)
            self.install_dir_edit.setEnabled(False)
            self.ui.shortcut_label.setEnabled(False)
            self.ui.shortcut_check.setEnabled(False)
        else:
            self.ui.shortcut_check.setChecked(QSettings().value("create_shortcut", True, bool))

        self.error_box()

        platforms = ["Windows"]
        if dl_item.options.app_name in self.api_results.bit32_games:
            platforms.append("Win32")
        if dl_item.options.app_name in self.api_results.mac_games:
            platforms.append("Mac")
        self.ui.platform_combo.addItems(platforms)
        self.ui.platform_combo.currentIndexChanged.connect(lambda: self.option_changed(None))
        self.ui.platform_combo.currentIndexChanged.connect(lambda: self.error_box())
        self.ui.platform_combo.currentIndexChanged.connect(
            lambda i: self.error_box(
                self.tr("Warning"),
                self.tr("You will not be able to run the game if you select <b>{}</b> as platform").format(
                    self.ui.platform_combo.itemText(i)
                ),
            )
            if (self.ui.platform_combo.currentText() == "Mac" and pf.system() != "Darwin")
            else None
        )
        self.ui.platform_combo.currentTextChanged.connect(self.setup_sdl_list)

        if pf.system() == "Darwin" and "Mac" in platforms:
            self.ui.platform_combo.setCurrentIndex(platforms.index("Mac"))

        self.advanced.ui.max_workers_spin.setValue(self.core.lgd.config.getint("Legendary", "max_workers", fallback=0))
        self.advanced.ui.max_workers_spin.valueChanged.connect(self.option_changed)

        self.advanced.ui.max_memory_spin.setValue(self.core.lgd.config.getint("Legendary", "max_memory", fallback=0))
        self.advanced.ui.max_memory_spin.valueChanged.connect(self.option_changed)

        self.advanced.ui.dl_optimizations_check.stateChanged.connect(self.option_changed)
        self.advanced.ui.force_download_check.stateChanged.connect(self.option_changed)
        self.advanced.ui.ignore_space_check.stateChanged.connect(self.option_changed)
        self.advanced.ui.download_only_check.stateChanged.connect(
            lambda: self.non_reload_option_changed("download_only")
        )
        self.ui.shortcut_check.stateChanged.connect(
            lambda: self.non_reload_option_changed("shortcut")
        )

        self.selectable_checks: List[TagCheckBox] = []
        self.config_tags: Optional[List[str]] = None
        self.setup_sdl_list("Mac" if pf.system() == "Darwin" and "Mac" in platforms else "Windows")

        self.ui.install_button.setEnabled(False)

        if self.dl_item.options.overlay:
            self.ui.platform_label.setVisible(False)
            self.ui.platform_combo.setVisible(False)
            self.advanced.ui.ignore_space_label.setVisible(False)
            self.advanced.ui.ignore_space_check.setVisible(False)
            self.advanced.ui.download_only_label.setVisible(False)
            self.advanced.ui.download_only_check.setVisible(False)
            self.ui.shortcut_label.setVisible(False)
            self.ui.shortcut_check.setVisible(False)
            self.selectable.setVisible(False)

        if pf.system() == "Darwin":
            self.ui.shortcut_check.setDisabled(True)
            self.ui.shortcut_check.setChecked(False)
            self.ui.shortcut_check.setToolTip(self.tr("Creating a shortcut is not supported on MacOS"))

        self.advanced.ui.install_prereqs_label.setEnabled(False)
        self.advanced.ui.install_prereqs_check.setEnabled(False)
        self.advanced.ui.install_prereqs_check.stateChanged.connect(
            lambda: self.non_reload_option_changed("install_prereqs")
        )

        self.non_reload_option_changed("shortcut")

        self.ui.cancel_button.clicked.connect(self.cancel_clicked)
        self.ui.verify_button.clicked.connect(self.verify_clicked)
        self.ui.install_button.clicked.connect(self.install_clicked)

        self.advanced.ui.install_prereqs_check.setChecked(self.dl_item.options.install_prereqs)

        self.ui.install_dialog_layout.setSizeConstraint(QLayout.SetFixedSize)

    def execute(self):
        if self.silent:
            self.reject_close = False
            self.get_download_info()
        else:
            self.verify_clicked()
            self.show()

    @pyqtSlot(str)
    def setup_sdl_list(self, platform="Windows"):
        for cb in self.selectable_checks:
            cb.disconnect()
            cb.deleteLater()
        self.selectable_checks.clear()

        if config_tags := self.core.lgd.config.get(self.game.app_name, 'install_tags', fallback=None):
            self.config_tags = config_tags.split(",")
        config_disable_sdl = self.core.lgd.config.getboolean(self.game.app_name, 'disable_sdl', fallback=False)
        sdl_name = get_sdl_appname(self.game.app_name)
        if not config_disable_sdl and sdl_name is not None:
            # FIXME: this should be updated whenever platform changes
            sdl_data = self.core.get_sdl_data(sdl_name, platform=platform)
            if sdl_data:
                widget = QWidget(self.selectable)
                layout = QVBoxLayout(widget)
                layout.setSpacing(0)
                for tag, info in sdl_data.items():
                    cb = TagCheckBox(info["name"], info["tags"])
                    if tag == "__required":
                        cb.setChecked(True)
                        cb.setDisabled(True)
                    if self.config_tags is not None:
                        if all(elem in self.config_tags for elem in info["tags"]):
                            cb.setChecked(True)
                    layout.addWidget(cb)
                    self.selectable_checks.append(cb)
                for cb in self.selectable_checks:
                    cb.stateChanged.connect(self.option_changed)
                self.selectable.setWidget(widget)
        else:
            self.selectable.setDisabled(True)

    def get_options(self):
        self.dl_item.options.base_path = self.install_dir_edit.text() if not self.update else None

        self.dl_item.options.max_workers = self.advanced.ui.max_workers_spin.value()
        self.dl_item.options.shared_memory = self.advanced.ui.max_memory_spin.value()
        self.dl_item.options.order_opt = self.advanced.ui.dl_optimizations_check.isChecked()
        self.dl_item.options.force = self.advanced.ui.force_download_check.isChecked()
        self.dl_item.options.ignore_space = self.advanced.ui.ignore_space_check.isChecked()
        self.dl_item.options.no_install = self.advanced.ui.download_only_check.isChecked()
        self.dl_item.options.platform = self.ui.platform_combo.currentText()
        self.dl_item.options.install_prereqs = self.advanced.ui.install_prereqs_check.isChecked()
        self.dl_item.options.create_shortcut = self.ui.shortcut_check.isChecked()
        if self.selectable_checks:
            self.dl_item.options.install_tag = [""]
            for cb in self.selectable_checks:
                if data := cb.isChecked():
                    # noinspection PyTypeChecker
                    self.dl_item.options.install_tag.extend(data)

    def get_download_info(self):
        self.dl_item.download = None
        info_worker = InstallInfoWorker(self.core, self.dl_item, self.game)
        info_worker.setAutoDelete(True)
        info_worker.signals.result.connect(self.on_worker_result)
        info_worker.signals.failed.connect(self.on_worker_failed)
        info_worker.signals.finished.connect(self.on_worker_finished)
        self.worker_running = True
        self.threadpool.start(info_worker)

    def verify_clicked(self):
        self.error_box()
        message = self.tr("Updating...")
        self.ui.download_size_text.setText(message)
        self.ui.download_size_text.setStyleSheet("font-style: italic; font-weight: normal")
        self.ui.install_size_text.setText(message)
        self.ui.install_size_text.setStyleSheet("font-style: italic; font-weight: normal")
        self.ui.cancel_button.setEnabled(False)
        self.ui.verify_button.setEnabled(False)
        self.ui.install_button.setEnabled(False)
        self.options_changed = False
        self.get_options()
        self.get_download_info()

    def option_changed(self, path) -> Tuple[bool, str, str]:
        self.options_changed = True
        self.ui.install_button.setEnabled(False)
        self.ui.verify_button.setEnabled(not self.worker_running)
        return True, path, ""

    def non_reload_option_changed(self, option: str):
        if option == "download_only":
            self.dl_item.options.no_install = self.advanced.ui.download_only_check.isChecked()
        elif option == "shortcut":
            QSettings().setValue("create_shortcut", self.ui.shortcut_check.isChecked())
            self.dl_item.options.create_shortcut = self.ui.shortcut_check.isChecked()
        elif option == "install_prereqs":
            self.dl_item.options.install_prereqs = self.advanced.ui.install_prereqs_check.isChecked()

    def cancel_clicked(self):
        if self.config_tags:
            config_helper.add_option(self.game.app_name, 'install_tags', ','.join(self.config_tags))
        else:
            # lk: this is purely for cleaning any install tags we might have added erroneously to the config
            config_helper.remove_option(self.game.app_name, 'install_tags')

        self.dl_item.download = None
        self.reject_close = False
        self.close()

    def install_clicked(self):
        self.reject_close = False
        self.close()

    def on_worker_result(self, dl_item: InstallDownloadModel):
        self.dl_item.download = dl_item
        download_size = self.dl_item.download.analysis.dl_size
        install_size = self.dl_item.download.analysis.install_size
        if download_size:
            self.ui.download_size_text.setText("{}".format(get_size(download_size)))
            self.ui.download_size_text.setStyleSheet("font-style: normal; font-weight: bold")
            self.ui.install_button.setEnabled(not self.options_changed)
        else:
            self.ui.install_size_text.setText(self.tr("Game already installed"))
            self.ui.install_size_text.setStyleSheet("font-style: italics; font-weight: normal")
        self.ui.install_size_text.setText("{}".format(get_size(install_size)))
        self.ui.install_size_text.setStyleSheet("font-style: normal; font-weight: bold")
        self.ui.verify_button.setEnabled(self.options_changed)
        self.ui.cancel_button.setEnabled(True)
        if pf.system() == "Windows" or ArgumentsSingleton().debug:
            if dl_item.igame.prereq_info and not dl_item.igame.prereq_info.get("installed", False):
                self.advanced.ui.install_prereqs_check.setEnabled(True)
                self.advanced.ui.install_prereqs_label.setEnabled(True)
                self.advanced.ui.install_prereqs_check.setChecked(True)
                prereq_name = dl_item.igame.prereq_info.get("name", "")
                prereq_path = os.path.split(dl_item.igame.prereq_info.get("path", ""))[-1]
                prereq_desc = prereq_name if prereq_name else prereq_path
                self.advanced.ui.install_prereqs_check.setText(
                    self.tr("Also install: {}").format(prereq_desc)
                )
        if self.silent:
            self.close()

    def on_worker_failed(self, message: str):
        error_text = self.tr("Error")
        self.ui.download_size_text.setText(error_text)
        self.ui.install_size_text.setText(error_text)
        self.error_box(error_text, message)
        self.ui.verify_button.setEnabled(self.options_changed)
        self.ui.cancel_button.setEnabled(True)
        if self.silent:
            self.show()

    def error_box(self, label: str = "", message: str = ""):
        self.ui.warning_label.setVisible(bool(label))
        self.ui.warning_label.setText(label)
        self.ui.warning_text.setVisible(bool(message))
        self.ui.warning_text.setText(message)

    def on_worker_finished(self):
        self.worker_running = False

    # lk: happens when close() is called, also when top right 'X' is pressed.
    # lk: reject any events not coming from the buttons in case the WM
    # lk: doesn't honor the window hints
    def closeEvent(self, a0: QCloseEvent) -> None:
        if self.reject_close:
            a0.ignore()
        else:
            self.threadpool.clear()
            self.threadpool.waitForDone()
            self.result_ready.emit(self.dl_item)
            super(InstallDialog, self).closeEvent(a0)

    def keyPressEvent(self, e: QKeyEvent) -> None:
        if e.key() == Qt.Key_Escape:
            self.cancel_clicked()


class InstallInfoWorker(QRunnable):
    class Signals(QObject):
        result = pyqtSignal(InstallDownloadModel)
        failed = pyqtSignal(str)
        finished = pyqtSignal()

    def __init__(self, core: LegendaryCore, dl_item: InstallQueueItemModel, game: Game = None):
        sys.excepthook = sys.__excepthook__
        super(InstallInfoWorker, self).__init__()
        self.signals = InstallInfoWorker.Signals()
        self.core = core
        self.dl_item = dl_item
        self.game = game

    @pyqtSlot()
    def run(self):
        try:
            if not self.dl_item.options.overlay:
                cli = LegendaryCLI(self.core)
                status = LgndrIndirectStatus()
                result = cli.install_game(
                    LgndrInstallGameArgs(**self.dl_item.options.as_install_kwargs(), indirect_status=status)
                )
                if result:
                    download = InstallDownloadModel(*result)
                else:
                    raise LgndrException(status.message)
            else:
                if not os.path.exists(path := self.dl_item.options.base_path):
                    os.makedirs(path)

                dlm, analysis, igame = self.core.prepare_overlay_install(
                    path=self.dl_item.options.base_path
                )

                download = InstallDownloadModel(
                    dlm=dlm,
                    analysis=analysis,
                    igame=igame,
                    game=self.game,
                    repair=False,
                    repair_file="",
                    res=ConditionCheckResult(),  # empty
                )

            if not download.res or not download.res.failures:
                self.signals.result.emit(download)
            else:
                self.signals.failed.emit("\n".join(str(i) for i in download.res.failures))
        except LgndrException as ret:
            self.signals.failed.emit(ret.message)
        except Exception as e:
            self.signals.failed.emit(str(e))
        self.signals.finished.emit()


class TagCheckBox(QCheckBox):
    def __init__(self, text, tags: List[str], parent=None):
        super(TagCheckBox, self).__init__(parent)
        self.setText(text)
        self.tags = tags

    def isChecked(self) -> Union[bool, List[str]]:
        return self.tags if super(TagCheckBox, self).isChecked() else False
